(******************************************************************************)
(* computer                                                                   *)
(* (C) 2010..2021 Steffen Solyga <solyga@gmx.de>                              *)
(******************************************************************************)
IMPLEMENTATION MODULE Comp;

IMPORT FIO;

CONST
  debug = FALSE;
  out ::= FIO.StandardOutput;
  err ::= FIO.ErrorOutput;
  dbg ::= FIO.ErrorOutput;
  vcMax = 32;

  acMax = 3;   (* 3 enough for l03 (no relations/conditions) *)

VAR
  args: ARRAY [ 1 .. acMax ] OF TValue;  (* "fake" stack *)
  ac: CARDINAL;
  vars: ARRAY [ 1 .. vcMax ] OF TValue;  (* "real" stack *)
  vc: CARDINAL;


PROCEDURE Err( s: ARRAY OF CHAR );
BEGIN
  FIO.WrStr( err, 'Comp: ' );
  FIO.WrStr( err, s );
  FIO.WrStr( err, '.' );
  FIO.WrLn( err );
  HALT;
END Err;

PROCEDURE ChkMax();
BEGIN
  IF ac > acMax THEN Err( 'Too many arguments' ) END;
END ChkMax;

PROCEDURE ChkMin( x: CARDINAL );
BEGIN
  IF ac < x THEN Err( 'Too little arguments' ) END;
END ChkMin;


PROCEDURE Val( v: TValue );
BEGIN
  IF debug THEN
    FIO.WrStr( dbg, 'Comp.Debug: ac ' );
    FIO.WrCard( dbg, ac, 0 );
    FIO.WrStr( dbg, ', vc ' );
    FIO.WrCard( dbg, vc, 0 );
    FIO.WrStr( dbg, ': Val ' );
    FIO.WrInt( dbg, v, 0 );
    FIO.WrLn( dbg );
  END;
  INC( ac ); ChkMax(); args[ ac ] := v;
END Val;

PROCEDURE Opr( o: TOperation );
BEGIN
  IF debug THEN
    FIO.WrStr( dbg, 'Comp.Debug: ac ' );
    FIO.WrCard( dbg, ac, 0 );
    FIO.WrStr( dbg, ', vc ' );
    FIO.WrCard( dbg, vc, 0 );
    FIO.WrStr( dbg, ': Opr ' );
    FIO.WrCard( dbg, ORD(o), 0 );
    FIO.WrLn( dbg );
  END;
  CASE o OF
  | oNeg: ChkMin(1); args[ac] := -args[ac];
  | oWrt: ChkMin(1); FIO.WrInt(out,args[ac],0); FIO.WrLn(out); DEC(ac);
  | oAdd: ChkMin(2); DEC(ac); args[ac] := args[ac] + args[ac+1];
  | oSub: ChkMin(2); DEC(ac); args[ac] := args[ac] - args[ac+1];
  | oMul: ChkMin(2); DEC(ac); args[ac] := args[ac] * args[ac+1];
  | oDiv: ChkMin(2); IF args[ac] = 0 THEN Err( 'Division by zero' ) END;
          DEC(ac); args[ac] := args[ac] DIV args[ac+1];
  | oNew: INC(vc); IF vc > vcMax THEN Err( 'Too many variables' ) END;
  ELSE
    Err( 'Bug: Invalid operation' );
  END;
END Opr;

PROCEDURE Sto( o: TOffset );
BEGIN
  IF debug THEN
    FIO.WrStr( dbg, 'Comp.Debug: ac ' );
    FIO.WrCard( dbg, ac, 0 );
    FIO.WrStr( dbg, ', vc ' );
    FIO.WrCard( dbg, vc, 0 );
    FIO.WrStr( dbg, ': Sto ' );
    FIO.WrCard( dbg, o, 0 );
    FIO.WrLn( dbg );
  END;
  ChkMin(1); vars[1+o] := args[ac]; DEC(ac);
END Sto;

PROCEDURE Lod( o: TOffset );
BEGIN
  IF debug THEN
    FIO.WrStr( dbg, 'Comp.Debug: ac ' );
    FIO.WrCard( dbg, ac, 0 );
    FIO.WrStr( dbg, ', vc ' );
    FIO.WrCard( dbg, vc, 0 );
    FIO.WrStr( dbg, ': Lod ' );
    FIO.WrCard( dbg, o, 0 );
    FIO.WrLn( dbg );
  END;
  INC(ac); ChkMax(); args[ac] := vars[1+o];
END Lod;

BEGIN
  ac := 0;
  vc := 0;
END Comp.
